#!/usr/bin/env python
# coding: utf-8
"""
Multi-line documentation placeholder if whole module needs to be explained.

Multi line documentation for data processing. Multi line documentation
for data processing. Multi line documentation for data processing.
Multi line documentation for data processing. Multi line documentation
for data processing. Multi line documentation for data processing.
Multi line documentation for data processing.
"""
import xarray as xr


def main(dataarray):
    """One line documentation for data processing.

    Multi line documentation for data processing. Multi line documentation
    for data processing. Multi line documentation for data processing. Multi
    line documentation for data processing. Multi line documentation for
    data processing. Multi line documentation for data processing. Multi
    line documentation for data processing. Multi line documentation for
    data processing.

    Arguments
    =========
    dataarray: <Description>
    argument1: <Description>
    ...
    argumentX: <Description>

    Returns
    =======
    dataarray: <Description>
    """
    assert isinstance(dataarray, xr.DataArray), "Input is not an xr.DataArray"

    # add data processing code here

    return dataarray



# -*- coding: utf-8 -*-
"""
Calculate temperature for adiabatic process in air.
"""
import xarray as xr
from imktk import constants as cs


def main(p, T_0, p_0):
    """Calculate temperature for adiabatic process in air.

    Arguments
    =========
    p: xr.DataArray
        pressure
    T_0: float
        Temperature of start condition in K
    p_0: float
        Pressure of start condition

    Returns
    =======
    dataarray: temperature for adiabatic process
    """
    assert isinstance(T_0, float), "Input is not a float"
    assert isinstance(p_0, float), "Input is not a float"
    assert isinstance(p, xr.DataArray), "Input is not a xr.DataArray"

    T_ad = T_0 * (p / p_0) ** ((cs.gamma_air - 1) / cs.gamma_air)
    return T_ad



import os
from .toolkit import _add_folder, _load_from_env
from . import tutorial


def get_version():
    try:
        # try new metadata package
        from importlib import metadata

        return metadata.version("imktk")
    except ImportError:
        # backup routine if Python version <= 3.7
        import pkg_resources

        return pkg_resources.get_distribution("imktk").version


__version__ = get_version()


def main():
    loc, da, ds = location()
    print(f"  Library version: {__version__}")
    print(f" Library location: {os.path.dirname(loc)}")
    print(f"Dataarray scripts: {da}")
    print(f"  Dataset scripts: {ds}")
    print(f"    Tutorial data: {tutorial._TUTORIAL_DATA.keys()}")


def location():
    loc = os.path.abspath(__file__)
    directory = os.path.dirname(loc)
    da = os.path.join(directory, "dataarray_methods")
    ds = os.path.join(directory, "dataset_methods")
    return (loc, da, ds)


_, da, ds = location()
_add_folder(da, "da")
_add_folder(ds, "ds")
_load_from_env()


if __name__ == "__main__":
    main()



"""
Physical constants used by several scripts.

This file allows for import and registration of all custom constants
in `_custom_defined.py`. Without this boilerplate code the auto_completion would
not know about the constants defined in the module.

Since `scipy` is already providing a bunch of
[constants](https://docs.scipy.org/doc/scipy/reference/constants.html) it is
not necessary to duplicate work. This constant module will first check if a
constant is defined in `_custom_defined.py`. If it is **not** defined there,
it will check scipy.constants` and finally throw an error if not found.
"""
from scipy import constants as _cons
from . import _custom_defined as _cd

__all__ = _cd.__all__ + _cons.__all__


def __getattr__(name):
    if name in _cd.__all__:
        return getattr(_cd, name)
    elif name in _cons.__all__:
        return getattr(_cons, name)
    else:
        raise AttributeError(f"Attribute {name} not found.")



"""
Physical constants used by several scripts.

"""
# fmt: off
__all__ = [
    "avogadro",
    "r_gas", "r_g",
    "c_p_air",
    "c_v_air",
    "gamma_air", "adiabatic_index_air"
]
# fmt: on

avogadro = 6.02214076e23  # mol-1
r_g = r_gas = 8.31446261815324  # J mol-1 K-1
c_p_air = 29.19  # J mol-1 K-1 for room conditions
c_v_air = 20.85  # J mol-1 K-1 for room conditions
gamma_air = adiabatic_index_air = c_p_air / c_v_air



#!/usr/bin/env python
# coding: utf-8
"""
Interpolation routine for flight track (incl. aircraft and satellites)

This script helps calculating interpolations along aircraft and satellite tracks.

Usage:

    ```python
    import imktk

    da = imktk.tutorial.open_dataset("flight")
    dims = dict(lat = [1.1,2.4], lon = [.21,.35])
    da.imktk.flight_track(**dims)
    ```
"""
import xarray as xr


def main(dataarray, name="flight", method="linear", interpolate_na=None, **dims):
    """Interpolation routine for flight track (incl. aircraft and satellites)

    This script helps calculating interpolations along aircraft and satellite tracks.

    Arguments
    =========
    dataarray: xr.DataArray to be mapped on the flight tracks
    name: Name of flight
    method: Interpolation method to be used
    dims: Dictionary of coord and list of values
    interpolate_na: Options for interpolation of nan values

    Inerpolation options are described here: https://docs.xarray.dev/en/stable/generated/xarray.DataArray.interpolate_na.html

    Returns
    =======
    result: xr.DataArray of interpolated values to flight track
    """
    # check if first argument is a dataarray
    assert isinstance(dataarray, xr.DataArray), "Input is not an xr.DataArray"

    # check if dims dictionary fits to the dataarray, based on ...
    msg = f"Some dim keys of {dims.keys()} not found in {list(dataarray.coords)}"
    # ... all keys are valid coordinate dimensions
    assert all([x in dataarray.coords for x in dims.keys()]), msg
    # ... flight tracks are lists
    assert all([isinstance(x, list) for x in dims.values()]), "Dimension values are not lists"
    # ... all flight tracks are the same length
    lengths = [len(x) for x in dims.values()]
    assert all([x == lengths[0] for x in lengths]), "Not all dims are of same length"

    if interpolate_na is not None:
        assert isinstance(interpolate_na, dict), "Interpolate options are not given as a dict"
        dataarray = dataarray.interpolate_na(**interpolate_na)

    criteria = {k: xr.DataArray(v, dims=name) for k, v in dims.items()}

    return dataarray.interp(**criteria, method=method)



# -*- coding: utf-8 -*-

"""
Calculate saturation vapour pressure over liquid water for temperatures 123 K < T < 332 K according to
Murphy & Koop, 2005 https://doi.org/10.1256/qj.04.94.
"""
import xarray as xr
import numpy as np
import warnings


def main(T):
    """Return saturation vapour pressure over liquid water in Pa for given Temperature in K.

    Arguments
    =========
    T: xr.DataArray
        in K

    Returns
    =======
    dataarray: saturation vapour pressure over liquid water in Pa.
    """
    assert isinstance(T, xr.DataArray), "Input is not an xr.DataArray"
    if (T.max(skipna=True) >= 332) | (T.min(skipna=True) <= 123):
        warnings.warn("At least one of the values of T is out of range.")
    p_liq = np.exp(
        54.842763
        - 6763.22 / T
        - 4.21 * np.log(T)
        + 0.000367 * T
        + np.tanh(0.0415 * (T - 218.8)) * (53.878 - 1331.22 / T - 9.44523 * np.log(T) + 0.014025 * T)
    )
    return p_liq



#!/usr/bin/env python
# coding: utf-8
"""
Calculate number density from mixing ratio, temperature and pressure within a xr.DataArray.
Usage:
    xr.DataArray.nd(temperature, pressure)

    temperature:   in K
    pressure:      in Pa
        either as vector in case of pressure levels or as matrix in case of model levels

    return value: number density in m-3

author: Stefan Versick (KIT)
"""
import xarray as xr
import numpy as np
from imktk import constants as cs


def main(dataarray, temp, press):
    """Return number density for dataarray."""
    assert isinstance(dataarray, xr.DataArray), "Input is not an xr.DataArray"
    assert isinstance(temp, xr.DataArray), "Temperature is not an xr.DataArray"
    assert isinstance(press, xr.DataArray), "Pressure is not an xr.DataArray"
    assert dataarray.shape == temp.shape, "Input and Temperature are not of the same shape"

    if press.ndim == 1:  # data on fixed pressure grid
        press_mdim = np.array(int(dataarray.size / press.size) * [press.values]).reshape(dataarray.shape)
    else:  # data on model levels
        press_mdim = press
    assert dataarray.shape == press_mdim.shape, "Input and Pressure are not of the same shape"
    var_nd = cs.avogadro * press_mdim / (cs.r_gas * temp) * dataarray
    return var_nd



#!/usr/bin/env python
# coding: utf-8
"""Calculate monthly anomalies within a xr.DataArray."""
import xarray as xr
from climatology import main as climatology


def main(dataarray):
    """Return monthly anomalies for dataarray.

    Arguments
    =========
    dataarray: Climate dataarray with `time` coordinate

    Returns
    =======
    dataarray: Normalized climate dataarray with `time` coordinate on months resolution
    """
    assert isinstance(dataarray, xr.DataArray), "Input is not an xr.DataArray"
    anomalies = dataarray.groupby("time.month") - climatology(dataarray)
    return anomalies



#!/usr/bin/env python
# coding=utf-8
"""Extensions for Xarray for analysing weather data."""
import functools
import importlib as ilib
import os
import sys
import xarray as xr
import logging

_DA_ENV = "IMKTK_DATAARRAY"
_DS_ENV = "IMKTK_DATASET"
_LOG_LEVEL = "IMKTK_LOGLEVEL"


def _get_log_level_via_env():
    loglevel = os.environ.get(_LOG_LEVEL, "ERROR")
    numeric_level = getattr(logging, loglevel.upper(), None)
    if not isinstance(numeric_level, int):
        raise ValueError(f"Invalid log level: {loglevel}")
    return numeric_level


logging.basicConfig(format="%(asctime)s - %(name)s - %(levelname)s - %(message)s", level=_get_log_level_via_env())
logger = logging.getLogger(__name__)


@xr.register_dataarray_accessor("imktk")
class _NyxDA(object):
    def __init__(self, data):
        self._data = data

    def _add_methods(self, folder):
        return _add_folder(folder, mode="da")

    def _load_env(self):
        return _load_from_env(mode="da")

    def __repr__(self):
        add_methods = [x for x in dir(self) if not x.startswith("_")]
        return f"IMKTK.Dataarray(scripts: {add_methods})"


@xr.register_dataset_accessor("imktk")
class _NyxDS(object):
    def __init__(self, data):
        self._data = data

    def _add_methods(self, folder):
        return _add_folder(folder, mode="ds")

    def _load_env(self):
        return _load_from_env(mode="ds")

    def __repr__(self):
        add_methods = [x for x in dir(self) if not x.startswith("_")]
        return f"IMKTK.Dataset(scripts: {add_methods})"


def _patch(func, funcname, mode):
    classobject = _NyxDA if mode == "da" else _NyxDS

    @functools.wraps(func)
    def method(accessor, *args, **kwargs):
        return func(accessor._data, *args, **kwargs)

    setattr(classobject, funcname, method)
    return func


def xtend_dataarray(da_folder=None):
    return _add_folder(da_folder, "da")


def xtend_dataset(ds_folder=None):
    return _add_folder(ds_folder, "ds")


def _add_folder(folder, mode=None):
    assert isinstance(mode, str) and mode.lower() in ["ds", "da"], "Can not understand mode {}".format(mode)
    assert os.path.isdir(folder), '"{}" is not a folder'.format(folder)
    mode = mode.lower()
    folder = os.path.realpath(folder)

    pythonfiles = [x[:-3] for x in os.listdir(folder) if x.endswith(".py") and not x.startswith("_")]
    if not pythonfiles:
        logger.warning("Folder %s is empty. No methods added.", folder)
        return None
    sys.path.insert(0, folder)
    for method in pythonfiles:
        try:
            lib = ilib.import_module(method, package="imktk")
            _patch(getattr(lib, "main"), method, mode)
            logger.info("Method: %s added in mode: %s", method, mode)
        except (SystemError, ImportError) as err:
            logger.error("Method: %s not loaded. Because: %s", method, err)
        except AttributeError as err:
            logger.error("Method: %s has no 'main' function (err: %s).", method, err)
    return True


def _load_from_env(mode=None):
    if mode is None:
        _ = [_load_from_env(x) for x in ["da", "ds"]]
        return None
    envvar = _DA_ENV if mode == "da" else _DS_ENV
    env = os.environ.get(envvar, False)
    if not env:
        logger.warning("ENV %s not set. No methods added.", env)
        return None
    folders = env.split(":")
    for folder in folders:
        logger.info("Adding folder %s as %s (ENV: %s)", folder, mode, env)
        _add_folder(folder, mode)



# -*- coding: utf-8 -*-

"""
Calculate saturation vapour pressure over ice for temperatures > 110 K according to
Murphy & Koop, 2005 https://doi.org/10.1256/qj.04.94.
"""
import xarray as xr
import numpy as np
import warnings


def main(T):
    """Return saturation vapour pressure over ice in Pa for given Temperature in K.

    Arguments
    =========
    T: xr.DataArray
        in K

    Returns
    =======
    dataarray: saturation vapour pressure over ice in Pa
    """
    assert isinstance(T, xr.DataArray), "Input is not a xr.DataArray"
    if (T.max(skipna=True) > 273.15) | (T.min(skipna=True) <= 110):
        warnings.warn("At least one of the values of T is out of range.")
    p_ice = np.exp(9.550426 - 5723.265 / T + 3.53068 * np.log(T) - 0.00728332 * T)
    return p_ice



""" Data Generation for tutorials and examples

This module provides artifical data for tutorials.
Each example data is provided via a function e.g. `_xarray_toy_weather_data`.
These functions are collected using the `_TUTORIAL_DATA` variable.
There the `key` is being defined by which the data can be called.
Finally, the method `open_dataset` actually delivers the data.

All functions have the same argument and parameter space for easy interoberability.
The current argument and parameter space is as follows:

```python
def function_name(seed):
    pass
```
"""
import numpy as np
import xarray as xr
import pandas as pd


def _xarray_flight_tracks():
    """Generate toy data for testing flight tracks."""
    return xr.DataArray(
        np.sin(0.3 * np.arange(100).reshape(5, 4, 5)),
        [("lat", np.arange(5)), ("lon", [0.1, 0.2, 0.3, 0.4]), ("lev", np.arange(5))],
    )


def _xarray_toy_weather_data(seed=None):
    """Generate toy weather data.

    Generate toy weather data using the method defined in
    https://xarray.pydata.org/en/stable/examples/weather-data.html#Toy-weather-data

    Arguments
    =========
    seed: Seed value for random number generation

    Returns
    =======
    dataset: Generated dataset
    """
    if seed:
        np.random.seed(seed)
    times = pd.date_range("2000-01-01", "2001-12-31", name="time")
    annual_cycle = np.sin(2 * np.pi * (times.dayofyear.values / 365.25 - 0.28))

    base = 10 + 15 * annual_cycle.reshape(-1, 1)
    tmin_values = base + 3 * np.random.randn(annual_cycle.size, 3)
    tmax_values = base + 10 + 3 * np.random.randn(annual_cycle.size, 3)

    dataset = xr.Dataset(
        {
            "tmin": (("time", "location"), tmin_values),
            "tmax": (("time", "location"), tmax_values),
        },
        {"time": times, "location": ["IA", "IN", "IL"]},
    )

    return dataset


# Dictionary for mapping keywords to each function providin example data
_TUTORIAL_DATA = {
    "toy_weather": _xarray_toy_weather_data,
    "xarray": _xarray_toy_weather_data,
    "flight": _xarray_flight_tracks,
}


def open_dataset(keyword, **kwargs):
    """Provides example data based on keyword.


    Arguments
    =========
    keyword: Keyword which identifies which method to use for data generation
    seed: Seed value for random number generation

    Returns
    =======
    dataset: Generated dataset
    """
    try:
        return _TUTORIAL_DATA[keyword](**kwargs)
    except KeyError:
        msg = f"Could not find {keyword} in tutorial data"
        raise KeyError(msg)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['imktk', 'imktk.constants', 'imktk.dataarray_methods', 'imktk.dataset_methods']

package_data = \
{'': ['*']}

install_requires = \
['netCDF4>=1.5.8,<2.0.0',
 'pandas<2',
 'scipy>=1.9.0,<2.0.0',
 'xarray>=0.20.1,<0.21.0']

entry_points = \
{'console_scripts': ['imktk = imktk:main']}

setup_kwargs = {
    'name': 'imktk',
    'version': '0.3.1',
    'description': 'Toolkit provided by IMK at KIT',
    'long_description': '# IMK Toolkit\n\nThis toolkit provides [post-processing scripts](/imktk) developed by members of the\n[Institute of Meteorology and Climate Research (IMK)](https://dev.to/epassaro/keep-your-research-reproducible-with-conda-pack-and-github-actions-339n)\nat the Karlsruhe Institute of Technology (KIT). The goal of this module is to\ngather together python post-processing scripts for the analysis of netCDF data\nand distribute them easily.\n\n## Usage\nSimply import the library using `import imktk`. From then on all scripts are\navailable using the `imktk` attribute:\n\n```python\nimport imktk\n\nds = imktk.tutorial.open_dataset("toy_weather")  # Load example dataset\nanomaly_free_tmin = ds.tmin.imktk.anomalies()  # Select dataarray `xr.tmin` and execute anomalies script\n```\n\nThe following is a list of available scripts:\n\n\n| Description | Example usage | Link to Script\n|--------|--------|-------------|\n|Calculate temperature for adiabatic process in air|`ds.tmin.imktk.ad_temp(temp_at_0, press_at_0)`| [here](./imktk/dataarray_methods/ad_temp.py)|\n|Calculate monthly anomalies|`ds.tmin.imktk.anomalies()`| [here](./imktk/dataarray_methods/anomalies.py)|\n|Calculate monthly climatology|`ds.tmin.imktk.climatology()`| [here](./imktk/dataarray_methods/climatology.py)|\n|Interpolation routine for flight track (incl. aircraft and satellites)| `ds.tmin.imktk.flight_track(**dims)` | [here](./imktk/dataarray_methods/flight_track.py) |\n|Calculate number density from mixing ratio, temperature and pressure| `ds.tmin.imktk.num_den(temp, press)` | [here](./imktk/dataarray_methods/num_den.py)|\n|Calculate saturation vapour pressure over ice for temperatures > 110 K| `ds.tmin.imktk.vapour_pres_ice()` | [here](./imktk/dataarray_methods/vapour_pres_ice.py)|\n|Calculate saturation vapour pressure over liquid water for temperatures 123 K < T < 332 K|`ds.tmin.imktk.vapour_pres_liq()` | [here](./imktk/dataarray_methods/vapour_pres_liq.py)|\n\n## Getting Started\n\nThe easiest method to test the module is to use an interactive session with docker.\nIn this environment you will have a Python 3 environment with all necessary dependencies already installed.\n\n```bash\ndocker run -it imktk/imktk:latest bash\n```\n\n> For the brave: You can test the latest release candidate by changing `latest` to `testing`\n\n## Install\n\nChoose one of the following methods to install the package:\n\n1. Install using `pip`\n2. Install using `conda`\n3. Install using `git clone`\n\n> This package supports only Python 3 with version `>=3.7`. If you are using\n> an earlier version of Python please consider updating.\n\n### `pip`\n\nReleases are automatically uploaded to PyPI. Please execute following command\nto install the package.\n\n```bash\npython3 -m pip install imktk\n```\n\n### `conda`\n\nCurrently the package does no support native installation using `conda`\nrespectively `conda-forge`. This feature is on the roadmap and you can follow\nits process using issue [#34](https://github.com/imk-toolkit/imk-toolkit/issues/34).\nThe current workaround for `conda` installation is to use the following steps\nfor any given environment `<env>`.\n\n1. Activate the environment\n\n    ```bash\n    conda activate <env>\n    ```\n\n2. Install using `pip`\n\n    ```bash\n    python3 -m pip install imktk\n    ```\n\n### `git clone`\n\nIt is also possible to install the package natively by cloning the repository.\nIf you are interested in using this method of installation please follow\nthese steps\n\n1. Install build dependencies\n\n    ```bash\n    python3 -m pip install build\n    ```\n\n2. Clone repository\n\n    ```bash\n    git clone https://github.com/imk-toolkit/imk-toolkit.git\n    ```\n\n3. Generate the Python packages\n\n    ```bash\n    python3 -m build  # or `make build`\n    ```\n\n4. Install packages\n\n    ```bash\n    pip3 install dist/imktk-<current.version>-py3-none-any.whl  # or `make install`\n    ```\n\n> Please be aware that this package uses `HDF5` and `netCDF` c-libraries in the\n> backend. If you are installing using `git clone` the `HDF5_DIR` environment\n> variable with the location of the HDF5 header files needs to be set.\n\n## Further reading\n\nIf you are interested in the inner workings of the package and details of the\nimplementation please refer to the embedded [README.md](/imktk/README.md).\n',
    'author': 'Uğur Çayoğlu',
    'author_email': 'Ugur.Cayoglu@kit.edu',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/imk-toolkit/imk-toolkit',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python
# coding: utf-8
"""Calculate monthly climatology within a xr.DataArray."""
import xarray as xr


def main(dataarray):
    """Return monthly climatology for dataarray."""
    assert isinstance(dataarray, xr.DataArray), "Input is not an xr.DataArray"
    clim = dataarray.groupby("time.month").mean("time")
    return clim
